import os
import signal
import time
from tempfile import NamedTemporaryFile

from django.core.management import call_command

from .. import models
from .. import utils
from .test_basic import NotifyTestBase


class CommandTest(NotifyTestBase):

    def tearDown(self):
        models._notification_type_cache = {}
        super().tearDown()

    def test_notifymail(self):

        utils.subscribe(self.user2_settings, self.TEST_KEY)
        utils.subscribe(self.user1_settings, self.TEST_KEY)
        utils.notify("This notification goes out by mail!", self.TEST_KEY)

        call_command("notifymail", cron=True)

        # No un-mailed notifications can be left!
        self.assertEqual(models.Notification.objects.filter(is_emailed=False).count(), 0)

        # Test that calling it again works but nothing gets sent
        call_command("notifymail", cron=True)

        # Now try the daemon
        pid_file = NamedTemporaryFile(delete=False)
        # Close it so its available for the other command
        pid_file.close()

        try:
            call_command('notifymail', daemon=True, pid=pid_file.name, no_sys_exit=False)
        except SystemExit:
            # It's normal for this command to exit
            pass

        with open(pid_file.name, 'r') as fp:
            pid = fp.read()
        os.unlink(pid_file.name)

        # Give it a second to start
        time.sleep(1)

        os.kill(int(pid), signal.SIGTERM)



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_nyt', '0003_subscription'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='subscription',
            field=models.ForeignKey(
                to_field='id',
                to='django_nyt.Subscription',
                blank=True,
                on_delete=django.db.models.deletion.SET_NULL,
                null=True),
            preserve_default=True,
        ),
    ]



from django.db import migrations, models
from django.utils import timezone


class Migration(migrations.Migration):

    dependencies = [
        ('django_nyt', '0006_auto_20141229_1630'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='modified',
            field=models.DateTimeField(auto_now=True, default=timezone.now()),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='settings',
            name='is_default',
            field=models.BooleanField(default=False, verbose_name='Default for new subscriptions'),
        ),
    ]



from django.contrib import admin
from django.utils.translation import gettext as _
from django_nyt import models, settings


class SettingsAdmin(admin.ModelAdmin):
    raw_id_fields = ('user',)
    list_display = ('user', 'interval',)


class SubscriptionAdmin(admin.ModelAdmin):
    raw_id_fields = ('settings',)
    list_display = ('display_user', 'notification_type', 'display_interval',)

    def display_user(self, instance):
        return instance.settings.user
    display_user.short_description = _("user")

    def display_interval(self, instance):
        return instance.settings.interval
    display_interval.short_description = _("interval")


class NotificationAdmin(admin.ModelAdmin):

    raw_id_fields = ('user', 'subscription')


if settings.ENABLE_ADMIN:
    admin.site.register(models.NotificationType)
    admin.site.register(models.Notification, NotificationAdmin)
    admin.site.register(models.Settings, SettingsAdmin)
    admin.site.register(models.Subscription, SubscriptionAdmin)



from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import reverse

from django_nyt import models
from django_nyt import utils


User = get_user_model()


class TestViews(TestCase):

    def setUp(self):
        super().setUp()
        self.TEST_KEY = 'test_key'
        self.user = User.objects.create_user(
            'lalala',
            password='password',
        )
        self.user_settings = models.Settings.get_default_setting(self.user)

    def tearDown(self):
        super().tearDown()
        models._notification_type_cache = {}

    def test_mark_read(self):
        utils.subscribe(self.user_settings, self.TEST_KEY)
        utils.notify("Test Is a Test", self.TEST_KEY)
        self.assertEqual(models.Notification.objects.filter(is_viewed=False).count(),
                         1)
        nid = models.Notification.objects.get().id
        self.client.login(username=self.user.username,
                          password='password')
        self.client.get(reverse('nyt:json_mark_read', args=(nid,)))
        self.assertEqual(models.Notification.objects.filter(is_viewed=False).count(),
                         0)



from django.urls import include, re_path as url

from . import views

app_name = 'nyt'

urlpatterns = [
    url(r'^json/get/$', views.get_notifications, name='json_get'),
    url(r'^json/get/(?P<latest_id>\d+)/$', views.get_notifications, name='json_get'),
    url(r'^json/mark-read/$', views.mark_read, name='json_mark_read_base'),
    url(r'^json/mark-read/(\d+)/$', views.mark_read, name='json_mark_read'),
    url(r'^json/mark-read/(?P<id_lte>\d+)/(?P<id_gte>\d+)/$', views.mark_read, name='json_mark_read'),
    url(r'^goto/(?P<notification_id>\d+)/$', views.goto, name='goto'),
    url(r'^goto/$', views.goto, name='goto_base'),
]


def get_pattern(app_name=app_name, namespace="nyt"):
    """Every url resolution takes place as "nyt:view_name".
       https://docs.djangoproject.com/en/dev/topics/http/urls/#topics-http-reversing-url-namespaces
    """
    import warnings
    warnings.warn(
        'django_nyt.urls.get_pattern is deprecated and will be removed in next version,'
        ' just use include(\'django_nyt.urls\')', DeprecationWarning
    )
    return include('django_nyt.urls')



from channels.routing import route

from . import consumers

channel_routing = [
    route("websocket.connect", consumers.ws_connect, path=r"^/nyt/?$"),
    route("websocket.disconnect", consumers.ws_disconnect),
    route("websocket.receive", consumers.ws_receive),
]



import logging

from channels import Group

from . import models, settings

logger = logging.getLogger(__name__)


def notify_subscribers(notifications, key):
    """
    Notify all open channels about new notifications
    """

    logger.debug("Broadcasting to subscribers")

    notification_type_ids = models.NotificationType.objects.values('key').filter(key=key)

    for notification_type in notification_type_ids:
        g = Group(
            settings.NOTIFICATION_CHANNEL.format(
                notification_key=notification_type['key']
            )
        )
        g.send(
            {'text': 'new-notification'}
        )



from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.db.models.signals import post_delete
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _

from django_nyt import settings

_notification_type_cache = {}


class NotificationType(models.Model):
    """
    Notification types are added on-the-fly by the
    applications adding new notifications
    """

    key = models.CharField(
        max_length=128,
        primary_key=True,
        verbose_name=_('unique key'),
        unique=True
    )

    # TODO: This isn't translatable
    label = models.CharField(
        max_length=128,
        verbose_name=_('optional label'),
        blank=True,
        null=True
    )

    content_type = models.ForeignKey(
        ContentType,
        blank=True,
        null=True,
        on_delete=models.SET_NULL
    )

    def __str__(self):
        return self.key

    class Meta:
        db_table = settings.DB_TABLE_PREFIX + '_notificationtype'
        verbose_name = _('type')
        verbose_name_plural = _('types')

    @classmethod
    def get_by_key(cls, key, content_type=None):
        if key in _notification_type_cache:
            return _notification_type_cache[key]
        try:
            nt = cls.objects.get(key=key)
        except cls.DoesNotExist:
            nt = cls.objects.create(key=key, content_type=content_type)
        _notification_type_cache[key] = nt
        return nt


@receiver([post_save, post_delete], sender=NotificationType)
def clear_notification_type_cache(*args, **kwargs):
    global _notification_type_cache
    _notification_type_cache = {}


class Settings(models.Model):
    """
    Reusable settings object for a subscription
    """

    user = models.ForeignKey(
        settings.USER_MODEL,
        on_delete=models.CASCADE,  # If a user is deleted, remove all settings
        verbose_name=_("user"),
        related_name='nyt_settings',
    )

    interval = models.SmallIntegerField(
        choices=settings.INTERVALS,
        verbose_name=_('interval'),
        default=settings.INTERVALS_DEFAULT,
    )

    is_default = models.BooleanField(
        default=False,
        verbose_name=_("Default for new subscriptions"),
    )

    def __str__(self):
        obj_name = _("Settings for %s") % getattr(
            self.user, self.user.USERNAME_FIELD)
        return obj_name

    class Meta:
        db_table = settings.DB_TABLE_PREFIX + '_settings'
        verbose_name = _('settings')
        verbose_name_plural = _('settings')

    def clean(self):
        if not self.is_default and self.pk and self.user:
            default_settings = Settings.objects.filter(
                user=self.user,
                is_default=True,
            ).exclude(pk=self.pk)
            if not default_settings.exists():
                raise ValidationError(
                    _("At minimum one default setting must exist for the user")
                )

    def save(self, *args, **kwargs):
        # We should check that it's the only default setting manually because
        # it's not possible to create a database constraint for this.
        # Instead of having a constraint, we unset all other is_default for
        # the user.
        if self.is_default:
            default_settings = Settings.objects.filter(
                user=self.user,
                is_default=True,
            ).exclude(pk=self.pk)
            default_settings.update(is_default=False)
        else:
            non_default_settings = Settings.objects.filter(
                user=self.user,
                is_default=False,
            ).exclude(pk=self.pk)
            if non_default_settings.exists():
                non_default_settings[0].is_default = True
                non_default_settings[0].save()
            else:
                raise ValueError("A user must have a default settings object")
        super(Settings, self).save(*args, **kwargs)

    @classmethod
    def get_default_setting(cls, user):
        return cls.objects.get_or_create(
            user=user,
            is_default=True
        )[0]


class Subscription(models.Model):

    # If settings are deleted, remove all subscriptions (CASCADE)
    settings = models.ForeignKey(
        Settings,
        verbose_name=_("settings"),
        on_delete=models.CASCADE,
    )

    notification_type = models.ForeignKey(
        NotificationType,
        verbose_name=_("notification type"),
        on_delete=models.CASCADE
    )

    object_id = models.CharField(
        max_length=64,
        null=True,
        blank=True,
        help_text=_('Leave this blank to subscribe to any kind of object'),
        verbose_name=_("object ID"),
    )

    send_emails = models.BooleanField(
        default=True,
        verbose_name=_("send emails"),
    )

    latest = models.ForeignKey(
        'Notification',
        null=True,
        blank=True,
        related_name='latest_for',
        verbose_name=_("latest notification"),
        on_delete=models.CASCADE,
    )

    def __str__(self):
        obj_name = _("Subscription for: %s") % (
            getattr(self.settings.user, self.settings.user.USERNAME_FIELD))
        return obj_name

    class Meta:
        db_table = settings.DB_TABLE_PREFIX + '_subscription'
        verbose_name = _('subscription')
        verbose_name_plural = _('subscriptions')


class Notification(models.Model):

    #: Either set the subscription
    subscription = models.ForeignKey(
        Subscription,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name=_("subscription"),
    )

    #: Or the user to receive the notification
    # If a user is deleted, remove all notifications (CASCADE)
    user = models.ForeignKey(
        settings.USER_MODEL,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        verbose_name=_("user"),
        related_name='nyt_notifications',
    )

    message = models.TextField()

    # TODO: Why 200?
    url = models.CharField(
        verbose_name=_('link for notification'),
        blank=True,
        null=True,
        max_length=200,
    )

    is_viewed = models.BooleanField(default=False)

    is_emailed = models.BooleanField(default=False)

    created = models.DateTimeField(auto_now_add=True)

    modified = models.DateTimeField(auto_now=True)

    occurrences = models.PositiveIntegerField(
        default=1,
        verbose_name=_('occurrences'),
        help_text=_(
            'If the same notification was fired multiple '
            'times with no intermediate notifications'
        ),
    )

    def save(self, *args, **kwargs):
        assert self.user or self.subscription
        if not self.user:
            self.user = self.subscription.settings.user
        super(Notification, self).save(*args, **kwargs)

    @classmethod
    def create_notifications(cls, key, **kwargs):
        """
        Creates notifications directly in database -- do not call directly,
        use django_nyt.notify(...)

        This is the old interface.
        """

        if not key or not isinstance(key, str):
            raise KeyError('No notification key (string) specified.')

        object_id = kwargs.pop('object_id', None)
        filter_exclude = kwargs.pop('filter_exclude', {})
        recipient_users = kwargs.pop('recipient_users', None)

        objects_created = []
        subscriptions = Subscription.objects.filter(
            notification_type__key=key
        ).exclude(
            **filter_exclude
        )
        if object_id:
            subscriptions = subscriptions.filter(
                Q(object_id=object_id) | Q(object_id=None)
            )
        if recipient_users:
            subscriptions = subscriptions.filter(
                settings__user__in=recipient_users,
            )

        subscriptions = subscriptions.prefetch_related('latest', 'settings')
        subscriptions = subscriptions.order_by('settings__user')
        prev_user = None

        for subscription in subscriptions:
            # Don't alert the same user several times even though overlapping
            # subscriptions occur.
            if subscription.settings.user == prev_user:
                continue

            # Check if it's the same as the previous message
            latest = subscription.latest
            if latest and (latest.message == kwargs.get('message', None)
                           and latest.url == kwargs.get('url', None)
                           and latest.is_viewed is False):
                # Both message and URL are the same, and it hasn't been viewed
                # so just increment occurrence count.
                latest.occurrences = latest.occurrences + 1
                latest.is_emailed = False
                latest.save()
            else:
                # Insert a new notification
                new_obj = cls.objects.create(
                    subscription=subscription,
                    **kwargs)
                objects_created.append(
                    new_obj
                )
                subscription.latest = new_obj
                subscription.save()
            prev_user = subscription.settings.user

        return objects_created

    def __str__(self):
        return "%s: %s" % (self.user, self.message)

    class Meta:
        db_table = settings.DB_TABLE_PREFIX + '_notification'
        verbose_name = _('notification')
        verbose_name_plural = _('notifications')
        ordering = ('-id',)



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_nyt', '0002_notification_settings'),
    ]

    operations = [
        migrations.CreateModel(
            name='Subscription', fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True, verbose_name='ID', serialize=False, primary_key=True
                    )
                ), (
                    'settings',
                    models.ForeignKey(
                        to_field='id', to='django_nyt.Settings', on_delete=models.CASCADE
                    )
                ), (
                    'notification_type',
                    models.ForeignKey(
                        to_field='key', to='django_nyt.NotificationType', on_delete=models.CASCADE
                    )
                ), (
                    'object_id',
                    models.CharField(
                        help_text='Leave this blank to subscribe to any kind of object',
                        null=True,
                        max_length=64,
                        blank=True
                    )
                ), (
                    'send_emails', models.BooleanField(
                        default=True
                    )
                ), (
                    'latest',
                    models.ForeignKey(
                        to_field='id',
                        to='django_nyt.Notification',
                        blank=True,
                        null=True,
                        on_delete=models.CASCADE
                    )
                ),
            ],
            options={
                'verbose_name': 'subscription', 'db_table': 'nyt_subscription', 'verbose_name_plural': 'subscriptions',
            },
            bases=(
                models.Model,
            ),
        ),
    ]



import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('django_nyt', '0004_notification_subscription'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='user',
            field=models.ForeignKey(verbose_name='user', to_field='id', blank=True, to=settings.AUTH_USER_MODEL, null=True, on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='subscription',
            name='latest',
            field=models.ForeignKey(verbose_name='latest notification', to_field='id', blank=True, to='django_nyt.Notification', null=True, on_delete=models.CASCADE),
        ),
        migrations.AlterField(
            model_name='settings',
            name='user',
            field=models.ForeignKey(to=settings.AUTH_USER_MODEL, to_field='id', verbose_name='user', on_delete=models.CASCADE),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='settings',
            field=models.ForeignKey(to='django_nyt.Settings', to_field='id', verbose_name='settings', on_delete=models.CASCADE),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='notification_type',
            field=models.ForeignKey(to='django_nyt.NotificationType', to_field='key', verbose_name='notification type', on_delete=models.CASCADE),
        ),
        migrations.AlterField(
            model_name='notificationtype',
            name='label',
            field=models.CharField(max_length=128, null=True, verbose_name='optional label', blank=True),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='object_id',
            field=models.CharField(help_text='Leave this blank to subscribe to any kind of object', max_length=64, null=True, verbose_name='object ID', blank=True),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='send_emails',
            field=models.BooleanField(default=True, verbose_name='send emails'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='subscription',
            field=models.ForeignKey(on_delete=django.db.models.deletion.SET_NULL, verbose_name='subscription', to_field='id', blank=True, to='django_nyt.Subscription', null=True),
        ),
    ]
